/*
 * cc.redberry.pipe: java library for implementation of concurrent pipelines
 * Copyright (c) 2010-2012
 * Bolotin Dmitriy <bolotin.dmitriy@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package cc.redberry.pipe;

/**
 * Base interface for passive object producer.
 *
 * @param <T> type of produced objects
 * @author Bolotin Dmitriy (bolotin.dmitriy@gmail.com)
 * @see cc.redberry.pipe.blocks.AbstractOutputPort
 */
public interface OutputPort<T> extends AutoCloseable {
    /**
     * Returns next element from this output port, or {@code null} if port is closed.
     *
     * @return next element from this output port, or {@code null} if port is closed
     * @throws InterruptedException
     */
    T take() throws InterruptedException;

    /**
     * Close current output port and free all used resources (threads, file streams, etc...).
     *
     * <p>Contract for implementation:<ul>
     *
     * <li>The implementation of this method must not throw an exceptions if port is already closed.</li>
     *
     * <li>{@link #take()} method of this port should return only {@code null}s after this method invocation.</li>
     *
     * </ul></p>
     *
     * <p>If the implementation of this method does not fulfill this contract, than methods {@link
     * cc.redberry.pipe.blocks.Merger#close()}, {@link cc.redberry.pipe.blocks.ParallelProcessor#close()}, etc... will
     * have an unpredictable behavior.</p>
     *
     * <p>Use {@link cc.redberry.pipe.blocks.AbstractOutputPort} for proper synchronization.</p>
     */
    @Override
    void close();
}
