/*
 * cc.redberry.pipe: java library for implementation of concurrent pipelines
 * Copyright (c) 2010-2012
 * Bolotin Dmitriy <bolotin.dmitriy@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package cc.redberry.pipe.blocks;


import cc.redberry.pipe.InputPort;
import cc.redberry.pipe.OutputPort;
import cc.redberry.pipe.util.UncaughtExceptionHandler;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

/**
 * Merges data streams form several output ports.
 *
 * <p>It starts a new thread for each attached output port to pull data from it and place it into output buffer.</p>
 *
 * @param <E> type of merged stream objects
 * @author Bolotin Dmitriy
 * @author Mikhail Shugay
 */
public final class Merger<E> implements OutputPort<E> {
    final Buffer<E> buffer;
    final List<OutputPort<? extends E>> inputPorts = new ArrayList<>();
    List<ExecutorService> targetExecutors = new ArrayList<>();
    final List<Future<?>> futures = new ArrayList<>();
    final List<Thread> threads = new ArrayList<>();
    private RuntimeException thrown = null;
    private final UncaughtExceptionHandler<O2ITransmitter> exceptionsHandler = new UncaughtExceptionHandler<O2ITransmitter>() {
        @Override
        public void handle(RuntimeException exception, O2ITransmitter source) {
            if (thrown == null)
                thrown = exception;
            //Close input port on exception
            source.getInput().close();
        }
    };

    /**
     * Creates merger with default output buffer size.
     */
    public Merger() {
        this(1024);
    }

    /**
     * Creates merger with defined output buffer size.
     *
     * @param bufferSize size of output buffer
     */
    public Merger(int bufferSize) {
        this.buffer = new Buffer<>(bufferSize);
    }

    /**
     * Attaches output port to merger.
     *
     * <p>Each port will be processed in the individual thread, so no concurrent calls to {@link
     * cc.redberry.pipe.OutputPort#take()} will be performed.</p>
     *
     * @param port            port to be served
     * @param executorService executor service that should be used to create new thread to serve provided output port
     */
    public void merge(OutputPort<E> port, ExecutorService executorService) {
        if (targetExecutors == null)
            throw new RuntimeException("This merger is already started.");
        inputPorts.add(port);
        targetExecutors.add(executorService);
    }

    /**
     * Attaches output port to merger.
     *
     * <p>Each port will be processed in the individual thread, so no concurrent calls to {@link
     * cc.redberry.pipe.OutputPort#take()} will be performed.</p>
     *
     * @param port port to be served
     */
    public void merge(OutputPort<E> port) {
        if (targetExecutors == null)
            throw new RuntimeException("This merger is already started.");
        inputPorts.add(port);
        targetExecutors.add(null);
    }

    /**
     * Starts threads for current merger
     */
    public void start() {
        if (targetExecutors == null)
            throw new RuntimeException("This merger is already started.");

        ExecutorService service;
        OutputPort<? extends E> port;
        O2ITransmitter<E> task;

        int i;

        //Preallocate buffers, because creation of new input port of Buffer
        //is forbidden after first "put" into it (through other input port)
        //TODO add exception in Buffer
        InputPort<E>[] iPorts = new InputPort[inputPorts.size()];
        for (i = 0; i < inputPorts.size(); ++i)
            iPorts[i] = buffer.createInputPort();

        for (i = 0; i < inputPorts.size(); ++i) {
            service = targetExecutors.get(i);
            port = inputPorts.get(i);
            task = new O2ITransmitter<>(port, iPorts[i], exceptionsHandler);
            if (service == null) {
                Thread thread = new Thread(task);
                threads.add(thread);
                thread.start();
            } else
                futures.add(service.submit(task));
        }
        targetExecutors = null;
    }

    /**
     * Join all threads owned by this buffer.
     *
     * @throws RuntimeException     if some of {@link cc.redberry.pipe.OutputPort#take()} thrown som exception
     * @throws InterruptedException
     */
    public void join() throws InterruptedException {
        try {
            for (Future<?> future : futures)
                future.get();
        } catch (ExecutionException ee) {
            throw new RuntimeException("This expression should not be thrown.", ee);
        }
        for (Thread thread : threads)
            thread.join();
        if (thrown != null)
            throw thrown;
        if (Thread.interrupted())
            throw new InterruptedException();
    }

    /**
     * Take object from output buffer
     *
     * @return {@inheritDoc}
     * @throws InterruptedException
     */
    @Override
    public E take() throws InterruptedException {
        return buffer.take();
    }

    /**
     * This method calls {@link #close(boolean)} method with {@code true} parameter.
     */
    @Override
    public void close() {
        close(true);
    }

    /**
     * Stops current merger. If {@code freeBuffer} parameter is {@code true} than output buffer of this merges will be
     * closed with {@code true} parameter (see {@link Buffer#close()}), so results in output
     * buffer will be dropped. If {@code freeBuffer} parameter is {@code false} all already processed and being
     * processed now items could be extracted through {@link #take()} method.
     *
     * <p>This method propagates close signal to upstream (input) output port.</p>
     *
     * @param freeBuffer indicates weather dropping of already processed elements is needed
     */
    public void close(boolean freeBuffer) {
        for (OutputPort op : inputPorts)
            op.close();

        //If each redirected input OutputPort comply with the contract for close, next take() will return null
        //in all threads. So no additional interrupt signaling is needed.

        /*for (Future<?> future : futures)
            future.cancel(true); //see O2ITransmitter code. All downstream output ports will be closed.

        for (Thread thread : threads)
            thread.interrupt(); //see O2ITransmitter code. All downstream output ports will be closed.

        try {
            join();
        } catch (InterruptedException ie) {
            throw new RuntimeException(ie);
        }*/

        if (freeBuffer)
            buffer.close(true);
    }
}
