package cc.redberry.pipe.blocks;

import cc.redberry.pipe.Processor;
import cc.redberry.pipe.ThreadSafe;

/**
 * {@link cc.redberry.pipe.Processor} wrapper used to save input value passed to the wrapped processor and attache it to
 * the output.
 *
 * @author Bolotin Dmitriy
 * @see InputOutputPair
 */
public class ResultAttachingProcessor<InputT, OutputT> implements Processor<InputT, InputOutputPair<InputT, OutputT>> {
    private final Processor<InputT, OutputT> processor;

    /**
     * Use factory method. {@link #create(cc.redberry.pipe.Processor)}.
     *
     * @param processor
     */
    protected ResultAttachingProcessor(Processor<InputT, OutputT> processor) {
        this.processor = processor;
    }

    @Override
    public final InputOutputPair<InputT, OutputT> process(InputT input) {
        OutputT output = processor.process(input);
        return new InputOutputPair<InputT, OutputT>(input, output);
    }

    public static final <InputT, OutputT> ResultAttachingProcessor<InputT, OutputT> create(Processor<InputT, OutputT> processor) {
        if (processor instanceof ThreadSafe)
            return new ThreadSafeProcessor<>(processor);
        else
            return new ResultAttachingProcessor<>(processor);
    }

    private static final class ThreadSafeProcessor<InputT, OutputT> extends ResultAttachingProcessor<InputT, OutputT> implements ThreadSafe {
        private ThreadSafeProcessor(Processor<InputT, OutputT> processor) {
            super(checked(processor));
        }

        private static final <InputT, OutputT> Processor<InputT, OutputT> checked(Processor<InputT, OutputT> processor) {
            if (!(processor instanceof ThreadSafe))
                throw new IllegalArgumentException();
            return processor;
        }
    }
}
